/*
Copyright Notice
================
BOCHS is Copyright 1994 by Kevin P. Lawton.

BOCHS is shareware for PERSONAL USE only.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@tiac.net
*/






#define BX_MAIN_C 1

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>


#include "bx_bochs.h"
#include "iodev/iodev.h"

/* prototypes */
void bx_usage(char *progname);
void bx_init_processor(void);

/* typedefs */


/* the device id and stepping id are loaded into DH & DL upon processor
   startup.  for device id: 3 = 80386, 4 = 80486.  just make up a
   number for the stepping (revision) id. */
#define BX_DEVICE_ID     2
#define BX_STEPPING_ID   0


bx_cpu_t bx_cpu;

extern Bit8u *bx_phy_memory; /* pointer to bochs physical memory space */
extern int bx_curr_instr_size; /* current instruction size */


FILE *test_fd;

static char command[128];

void bx_sanity_checks();


  int
main(int argc, char *argv[])
{
  char ch;
  char *ptr;
  int n;


  if (argc < 2) bx_usage(argv[0]);
  n = 2;
  while (n <= argc) {
    if (n!=argc && !strcmp(argv[n-1], "-script")) {
      test_fd = fopen(argv[n], "r");
      if (!test_fd) {
	fprintf(stderr, "could not open test file\n");
	exit(1);
	}
      n += 2;
      }
    else if (!strcmp(argv[n-1], "-sanity-check")) {
      bx_sanity_checks();
      exit(0);
      n += 1;
      }
    else {
      bx_usage(argv[0]);
      }
    }


  bx_init_processor();
  bx_iodev_setup();


  while (1) {
    fscanf(test_fd, "%s", command);
    if (feof(test_fd)) exit(0); /* all done reading script file */
    if (command[0] == '#') {
      while ( (ch = fgetc(test_fd)) != '\n' && !feof(test_fd) );
      }
    else if (!strcmp(command, ".exit")) {
      fprintf(stderr, "\n");
      exit(0);
      }
    else if (!strcmp(command, ".i")) {
      /*fprintf(stderr, "\n>>> decode()\n");*/
      bx_decode();
      /* probably check if in 16 or 32 bit mode, but incrementing
         the instruction pointer always in 32 bit mode should be OK */
      bx_cpu.eip += bx_curr_instr_size;
      }
    else if (!strcmp(command, ".print")) {
      while (1) {
        strcpy(command, (char *) "   ");
        ch = fgetc(test_fd);
        if (ch == '\n') {
          fputc('\n', stderr);
          break;
          }
        ungetc(ch, test_fd);
	if (fscanf(test_fd, "%s", command) != 1) {
          fprintf(stderr, "Error: bad print command\n");
          exit(0);
          }

	ptr = command;
	while (*ptr)
          if (isupper(*ptr))
	    *ptr++ = tolower(*ptr);
          else ptr++;
	if (!strcmp(command, "eax"))
	  fprintf(stderr, "EAX:    %08X", (unsigned) bx_cpu.eax);
	else if (!strcmp(command, "edx"))
	  fprintf(stderr, "EDX:    %08X", (unsigned) bx_cpu.edx);
	else if (!strcmp(command, "ecx"))
	  fprintf(stderr, "ECX:    %08X", (unsigned) bx_cpu.ecx);
	else if (!strcmp(command, "ebx"))
	  fprintf(stderr, "EBX:    %08X", (unsigned) bx_cpu.ebx);
	else if (!strcmp(command, "ebp"))
	  fprintf(stderr, "EBP:    %08X", (unsigned) bx_cpu.ebp);
	else if (!strcmp(command, "esi"))
	  fprintf(stderr, "ESI:    %08X", (unsigned) bx_cpu.esi);
	else if (!strcmp(command, "edi"))
	  fprintf(stderr, "EDI:    %08X", (unsigned) bx_cpu.edi);
	else if (!strcmp(command, "esp"))
	  fprintf(stderr, "ESP:    %08X", (unsigned) bx_cpu.esp);
	else if (!strcmp(command, "eflags")) {
	  fprintf(stderr, "EFLAGS:\n");
	  fprintf(stderr, "  CF:   %d\n", (unsigned) bx_cpu.eflags.cf);
	  fprintf(stderr, "  PF:   %d\n", (unsigned) bx_cpu.eflags.pf);
	  fprintf(stderr, "  AF:   %d\n", (unsigned) bx_cpu.eflags.af);
	  fprintf(stderr, "  ZF:   %d\n", (unsigned) bx_cpu.eflags.zf);
	  fprintf(stderr, "  SF:   %d\n", (unsigned) bx_cpu.eflags.sf);
	  fprintf(stderr, "  TF:   %d\n", (unsigned) bx_cpu.eflags.tf);
	  fprintf(stderr, "  IF:   %d\n", (unsigned) bx_cpu.eflags.if_);
	  fprintf(stderr, "  DF:   %d\n", (unsigned) bx_cpu.eflags.df);
	  fprintf(stderr, "  OF:   %d\n", (unsigned) bx_cpu.eflags.of);
	  fprintf(stderr, "  IOPL: %d\n", (unsigned) bx_cpu.eflags.iopl);
	  fprintf(stderr, "  NT:   %d\n", (unsigned) bx_cpu.eflags.nt);
	  fprintf(stderr, "  RF:   %d\n", (unsigned) bx_cpu.eflags.rf);
	  fprintf(stderr, "  VM:   %d\n", (unsigned) bx_cpu.eflags.vm);
	  fprintf(stderr, "  AC:   %d\n", (unsigned) bx_cpu.eflags.ac);
	  }
	else if (!strcmp(command, "eip"))
	  fprintf(stderr, "EIP:    %08X", (unsigned) bx_cpu.eip);
	else if (!strcmp(command, "cs"))
	  fprintf(stderr, "cs:     %04X", (unsigned) bx_cpu.cs.selector.value);
	else if (!strcmp(command, "ds"))
	  fprintf(stderr, "ds:     %04X", (unsigned) bx_cpu.ds.selector.value);
	else if (!strcmp(command, "ss"))
	  fprintf(stderr, "ss:     %04X", (unsigned) bx_cpu.ss.selector.value);
	else if (!strcmp(command, "es"))
	  fprintf(stderr, "es:     %04X", (unsigned) bx_cpu.es.selector.value);
	else if (!strcmp(command, "fs"))
	  fprintf(stderr, "fs:     %04X", (unsigned) bx_cpu.fs.selector.value);
	else if (!strcmp(command, "gs"))
	  fprintf(stderr, "gs:     %04X", (unsigned) bx_cpu.gs.selector.value);
	else {
	  fprintf(stderr, "%s", command);
          while ( (ch = fgetc(test_fd)) != '\n')
            fputc(ch, stderr);
          fputc('\n', stderr);
          break;
	  }
        while ( (ch = fgetc(test_fd)) == ' ');
        if (ch == '\n') {
          fputc('\n', stderr);
          break;
	  }
        else
          ungetc(ch, test_fd);
        fputc('\n', stderr);
        } /* while (1) */
      } /* else if .print */

    else if (!strcmp(command, ".set")) {
      while (1) {
	if (fscanf(test_fd, "%s", command) != 1) {
          fprintf(stderr, "!!! bad .set command\n");
          exit(0);
          }
	ptr = command;
	while (*ptr)
          if (isupper(*ptr))
	    *ptr++ = tolower(*ptr);
          else
            ptr++;
	if (!strcmp(command, "eax"))
	  bx_cpu.eax = bx_fetch_next_dword();
	else if (!strcmp(command, "edx"))
	  bx_cpu.edx = bx_fetch_next_dword();
	else if (!strcmp(command, "ecx"))
	  bx_cpu.ecx = bx_fetch_next_dword();
	else if (!strcmp(command, "ebx"))
	  bx_cpu.ebx = bx_fetch_next_dword();
	else if (!strcmp(command, "ebp"))
	  bx_cpu.ebp = bx_fetch_next_dword();
	else if (!strcmp(command, "esi"))
	  bx_cpu.esi = bx_fetch_next_dword();
	else if (!strcmp(command, "edi"))
	  bx_cpu.edi = bx_fetch_next_dword();
	else if (!strcmp(command, "esp"))
	  bx_cpu.esp = bx_fetch_next_dword();
	else if (!strcmp(command, "eflags"))
	  bx_write_eflags(bx_fetch_next_dword());
	else if (!strcmp(command, "eip"))
	  bx_cpu.eip = bx_fetch_next_dword();
	else if (!strcmp(command, "cs"))
	  bx_cpu.cs.selector.value = bx_fetch_next_word();
	else if (!strcmp(command, "ds"))
	  bx_cpu.ds.selector.value = bx_fetch_next_word();
	else if (!strcmp(command, "ss"))
	  bx_cpu.ss.selector.value = bx_fetch_next_word();
	else if (!strcmp(command, "es"))
	  bx_cpu.es.selector.value = bx_fetch_next_word();
	else if (!strcmp(command, "fs"))
	  bx_cpu.fs.selector.value = bx_fetch_next_word();
	else if (!strcmp(command, "gs"))
	  bx_cpu.gs.selector.value = bx_fetch_next_word();
	else {
	  fprintf(stderr, "!!! unrecognized reg in set command\n");
	  exit(0);
	  }
        while ( (ch = fgetc(test_fd)) == ' ');
        if (ch == '\n')
          break;
        else
          ungetc(ch, test_fd);
        } /* while (1) */
      } /* else if .set */

    else {
      fprintf(stderr, "\n!!! Not handled: %s\n", command);
      }
    }
}

  void
bx_usage(char *progname)
{
  fprintf(stderr, "Usage: %s [-sanity-check] [-script testfile]\n", progname);
  exit(-1);
}

  void
bx_invalid_instruction(void)
{
  bx_error(1, "Invalid instruction\n");
}

  Bit8u
bx_fetch_next_byte(void)
{
  char temp[64];
  int next_byte;
  int ret;
  int i, bitcount;

  ret = fscanf(test_fd, "%s", temp);
  if (ret < 1) {
    fprintf(stderr, "bx_fetch_next_byte(): could not get next byte!\n");
    exit(0);
    }
  /*fprintf(stderr, "%s ", temp);*/
  if (temp[0] == '%') {
    next_byte = 0;
    for (i=1, bitcount=0; bitcount<8  && temp[i]; i++) {
      if (temp[i] != '0'  &&  temp[i] != '1'  && temp[i] != '-') {
        fprintf(stderr, "bx_fetch_next_byte(): error in binary constant!\n");
        exit(0);
        }
      if (temp[i] == '1') {
        next_byte = (next_byte << 1) + 1;
        bitcount++;
        }
      else if (temp[i] == '0') {
        next_byte = next_byte << 1;
        bitcount++;
        }
      }
    if (bitcount != 8) {
      fprintf(stderr, "bx_fetch_next_byte(): error in binary constant!\n");
      exit(0);
      }
    }
  else {
    next_byte = 0;
    sscanf(temp, "%x", &next_byte);
    }
  assert(next_byte < 256);
  bx_curr_instr_size++;
  return((Bit8u) next_byte);
}

  Bit32u
bx_fetch_next_dword(void)
{
  Bit8u b1, b2, b3, b4;
  Bit32u sum;

  b1 = bx_fetch_next_byte();
  b2 = bx_fetch_next_byte();
  b3 = bx_fetch_next_byte();
  b4 = bx_fetch_next_byte();
  sum = (((Bit32u)b4)<<24) + (((Bit32u)b3)<<16) +
    (((Bit32u)b2)<<8) + ((Bit32u)b1);
  return(sum);
}

  Bit16u
bx_fetch_next_word(void)
{
  Bit8u b1, b2;
  Bit16u sum;

  b1 = bx_fetch_next_byte();
  b2 = bx_fetch_next_byte();
  sum = (((Bit16u)b2)<<8) + ((Bit16u)b1);
  return(sum);
}

  void
bx_init_processor(void)
{
  /* general registers */
  bx_cpu.eax = 0; /* processor passed test :-) */
  bx_cpu.ebx = 0; /* undefined */
  bx_cpu.ecx = 0; /* undefined */
  bx_cpu.edx = (BX_DEVICE_ID << 8) | BX_STEPPING_ID;
  bx_cpu.ebp = 0; /* undefined */
  bx_cpu.esi = 0; /* undefined */
  bx_cpu.edi = 0; /* undefined */
  bx_cpu.esp = 0; /* undefined */

  /* status and control flags register set */
  bx_cpu.eflags.cf = 0;
  bx_cpu.eflags.pf = 0;
  bx_cpu.eflags.af = 0;
  bx_cpu.eflags.zf = 0;
  bx_cpu.eflags.sf = 0;
  bx_cpu.eflags.tf = 0;
  bx_cpu.eflags.if_ = 0;
  bx_cpu.eflags.df = 0;
  bx_cpu.eflags.of = 0;
  bx_cpu.eflags.iopl = 0;
  bx_cpu.eflags.nt = 0;
  bx_cpu.eflags.rf = 0;
  bx_cpu.eflags.vm = 0;
  bx_cpu.eflags.ac = 0;

  bx_cpu.eip    = 0x0000FFF0; /* instruction pointer */

  /* segment registers & descriptor caches */

  bx_cpu.cs.selector.value =     0xF000;
  bx_cpu.cs.selector.index =     0xF000;
  bx_cpu.cs.base           = 0xFFFF0000;
  bx_cpu.cs.limit          =     0xFFFF;
  bx_cpu.cs.attributes.d   =          0;
  bx_cpu.cs.attributes.dpl =          0;

  bx_cpu.ds.selector.value =     0x0000;
  bx_cpu.ds.selector.index =     0x0000;
  bx_cpu.ds.base           = 0x00000000;
  bx_cpu.ds.limit          =     0xFFFF;
  bx_cpu.ds.attributes.d   =          0;
  bx_cpu.ds.attributes.dpl =          0;

  bx_cpu.ss.selector.value =     0x0000;
  bx_cpu.ss.selector.index =     0x0000;
  bx_cpu.ss.base           = 0x00000000;
  bx_cpu.ss.limit          =     0xFFFF;
  bx_cpu.ss.attributes.d   =          0;
  bx_cpu.ss.attributes.dpl =          0;

  bx_cpu.es.selector.value =     0x0000;
  bx_cpu.es.selector.index =     0x0000;
  bx_cpu.es.base           = 0x00000000;
  bx_cpu.es.limit          =     0xFFFF;
  bx_cpu.es.attributes.d   =          0;
  bx_cpu.es.attributes.dpl =          0;

  bx_cpu.fs.selector.index =     0x0000;
  bx_cpu.fs.selector.index =     0x0000;
  bx_cpu.fs.base           = 0x00000000;
  bx_cpu.fs.limit          =     0xFFFF;
  bx_cpu.fs.attributes.d   =          0;
  bx_cpu.fs.attributes.dpl =          0;

  bx_cpu.gs.selector.index =     0x0000;
  bx_cpu.gs.selector.index =     0x0000;
  bx_cpu.gs.base           = 0x00000000;
  bx_cpu.gs.limit          =     0xFFFF;
  bx_cpu.gs.attributes.d   =          0;
  bx_cpu.gs.attributes.dpl =          0;

  bx_cpu.gdtr.base         = 0x00000000;  /* undefined */
  bx_cpu.gdtr.limit        =     0x0000;  /* undefined */

  bx_cpu.idtr.base         = 0x00000000;
  bx_cpu.idtr.limit        =     0x03FF; /* always byte granular */

  bx_cpu.ldtr.selector.index =     0x0000; /* undefined */
  bx_cpu.ldtr.base           = 0x00000000; /* undefined */
  bx_cpu.ldtr.limit          =     0x0000; /* undefined */

  bx_cpu.tr.selector.index   =     0x0000; /* undefined */
  bx_cpu.tr.base             = 0x00000000; /* undefined */
  bx_cpu.tr.limit            =     0x0000; /* undefined */

  bx_cpu.cr0.pg = 0; /* disable paging */
  bx_cpu.cr0.cd = 1; /* caching disabled */
  bx_cpu.cr0.nw = 1; /* no write-through */
  bx_cpu.cr0.am = 0; /* disable alignment check */
  bx_cpu.cr0.wp = 0; /* disable write-protect */
  bx_cpu.cr0.ne = 0; /* ndp exceptions through int 13H, DOS compat */
  bx_cpu.cr0.et = 0; /* coprocessor not present (for now) */
  bx_cpu.cr0.ts = 0; /* no task switch */
  bx_cpu.cr0.em = 1; /* emulate math coprocessor */
  bx_cpu.cr0.mp = 0; /* wait instructions not trapped */
  bx_cpu.cr0.pe = 0; /* real mode */


  /* debug registers (unimplemented) */
  bx_cpu.dr0 = 0;   /* undefined */
  bx_cpu.dr1 = 0;   /* undefined */
  bx_cpu.dr2 = 0;   /* undefined */
  bx_cpu.dr3 = 0;   /* undefined */
  bx_cpu.dr4 = 0;   /* undefined */
  bx_cpu.dr5 = 0;   /* undefined */
  bx_cpu.dr6 = 0xFFFF0FF0;   /* unimplemented */
  bx_cpu.dr7 = 0x00000000;   /* unimplemented */

  /* test registers 3-7 (unimplemented) */
  bx_cpu.tr3 = 0;   /* undefined */
  bx_cpu.tr4 = 0;   /* undefined */
  bx_cpu.tr5 = 0;   /* undefined */
  bx_cpu.tr6 = 0;   /* undefined */
  bx_cpu.tr7 = 0;   /* undefined */

  bx_init_memory(BX_PHY_MEMSIZE);
}


  void
bx_sanity_checks()
{
  Bit8u al, cl, dl, bl, ah, ch, dh, bh;
  Bit16u ax, cx, dx, bx, sp, bp, si, di;
  Bit32u eax, ecx, edx, ebx, esp, ebp, esi, edi;

  bx_cpu.eax = 0xFFEEDDCC;
  bx_cpu.ecx = 0xBBAA9988;
  bx_cpu.edx = 0x77665544;
  bx_cpu.ebx = 0x332211FF;
  bx_cpu.esp = 0xEEDDCCBB;
  bx_cpu.ebp = 0xAA998877;
  bx_cpu.esi = 0x66554433;
  bx_cpu.edi = 0x2211FFEE;

  BX_READ_8BIT_REG(al, BX_8BIT_REG_AL);
  BX_READ_8BIT_REG(cl, BX_8BIT_REG_CL);
  BX_READ_8BIT_REG(dl, BX_8BIT_REG_DL);
  BX_READ_8BIT_REG(bl, BX_8BIT_REG_BL);
  BX_READ_8BIT_REG(ah, BX_8BIT_REG_AH);
  BX_READ_8BIT_REG(ch, BX_8BIT_REG_CH);
  BX_READ_8BIT_REG(dh, BX_8BIT_REG_DH);
  BX_READ_8BIT_REG(bh, BX_8BIT_REG_BH);
  if ( al != (bx_cpu.eax & 0xFF) ||
       cl != (bx_cpu.ecx & 0xFF) ||
       dl != (bx_cpu.edx & 0xFF) ||
       bl != (bx_cpu.ebx & 0xFF) ||
       ah != ((bx_cpu.eax >> 8) & 0xFF) ||
       ch != ((bx_cpu.ecx >> 8) & 0xFF) ||
       dh != ((bx_cpu.edx >> 8) & 0xFF) ||
       bh != ((bx_cpu.ebx >> 8) & 0xFF) ) {
    bx_error(1, "problems using BX_READ_8BIT_REG()!\n");
    }

  BX_READ_16BIT_REG(ax, BX_16BIT_REG_AX);
  BX_READ_16BIT_REG(cx, BX_16BIT_REG_CX);
  BX_READ_16BIT_REG(dx, BX_16BIT_REG_DX);
  BX_READ_16BIT_REG(bx, BX_16BIT_REG_BX);
  BX_READ_16BIT_REG(sp, BX_16BIT_REG_SP);
  BX_READ_16BIT_REG(bp, BX_16BIT_REG_BP);
  BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
  BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
  if ( ax != (bx_cpu.eax & 0xFFFF) ||
       cx != (bx_cpu.ecx & 0xFFFF) ||
       dx != (bx_cpu.edx & 0xFFFF) ||
       bx != (bx_cpu.ebx & 0xFFFF) ||
       sp != (bx_cpu.esp & 0xFFFF) ||
       bp != (bx_cpu.ebp & 0xFFFF) ||
       si != (bx_cpu.esi & 0xFFFF) ||
       di != (bx_cpu.edi & 0xFFFF) ) {
    bx_error(1, "problems using BX_READ_16BIT_REG()!\n");
    }


  BX_READ_32BIT_REG(eax, BX_32BIT_REG_EAX);
  BX_READ_32BIT_REG(ecx, BX_32BIT_REG_ECX);
  BX_READ_32BIT_REG(edx, BX_32BIT_REG_EDX);
  BX_READ_32BIT_REG(ebx, BX_32BIT_REG_EBX);
  BX_READ_32BIT_REG(esp, BX_32BIT_REG_ESP);
  BX_READ_32BIT_REG(ebp, BX_32BIT_REG_EBP);
  BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
  BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
  if ( eax != bx_cpu.eax ||
       ecx != bx_cpu.ecx ||
       edx != bx_cpu.edx ||
       ebx != bx_cpu.ebx ||
       esp != bx_cpu.esp ||
       ebp != bx_cpu.ebp ||
       esi != bx_cpu.esi ||
       edi != bx_cpu.edi ) {
    bx_error(1, "problems using BX_READ_32BIT_REG()!\n");
    }


  if (sizeof(Bit8u)  != 1  ||  sizeof(Bit8s)  != 1)
    bx_error(1, "data type Bit8u or Bit8s is not of length 1 byte!\n");
  if (sizeof(Bit16u) != 2  ||  sizeof(Bit16s) != 2)
    bx_error(1, "data type Bit16u or Bit16s is not of length 2 bytes!\n");
  if (sizeof(Bit32u) != 4  ||  sizeof(Bit32s) != 4)
    bx_error(1, "data type Bit32u or Bit32s is not of length 4 bytes!\n");
  
  fprintf(stderr, "all sanity checks passed!\n");
}
