/*
Copyright Notice
================
BOCHS is Copyright 1994 by Kevin P. Lawton.

BOCHS is shareware for PERSONAL USE only.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@tiac.net
*/






#include <stdio.h>
#include <assert.h>
#include "bx_bochs.h"
#include "stdlib.h"


#define BX_PAGE_SIZE               0x1000  /* 4 Kbyte pages */
#define BX_NUM_PAGES               0x100   /* 256 pages */
                                           /* 4K * 256 = 1M */

#define BX_PAGE_UNINITIALIZED 0x01
#define BX_PAGE_IO_MAPPED     0x02

typedef struct {
  Bit8u *mem;
  Bit8u flags;
  } bx_phy_mem_page_t;


bx_phy_mem_page_t bx_phy_mem_pages[BX_NUM_PAGES];



char *bx_strseg(bx_segment_reg_t *seg);
Bit32u bx_translate_linear(Bit32u addr, Bit8u pl, Bit8u rw);
void bx_access_physical(Bit32u addr, Bit8u len, Bit8u rw, void *data);



  char *
bx_strseg(bx_segment_reg_t *seg)
{
  if (seg == &bx_cpu.es) return("ES");
  else if (seg == &bx_cpu.cs) return("CS");
  else if (seg == &bx_cpu.ss) return("SS");
  else if (seg == &bx_cpu.ds) return("DS");
  else if (seg == &bx_cpu.fs) return("FS");
  else if (seg == &bx_cpu.gs) return("GS");
  else return("??");
}

  void
bx_init_memory(int memsize)
{
  int i;

  for (i=0; i<BX_NUM_PAGES; i++) {
    bx_phy_mem_pages[i].mem = NULL;
    bx_phy_mem_pages[i].flags = BX_PAGE_UNINITIALIZED;
    }
}

  void
bx_access_virtual(bx_segment_reg_t *seg, Bit32u offset, Bit8u length, 
    Bit8u rw, void *data)
{
  fprintf(stderr, "%-5s access to vm:%s[%u = %08x]*%u\n",
    (rw==BX_READ)?"read":"write", bx_strseg(seg),
    (unsigned) offset, (unsigned) offset, (unsigned) length);

  /* if real or virtual 8086 mode */
  if (!bx_cpu.cr0.pe) {
    if (offset + length  >  0x10000) {
      if (seg == &bx_cpu.ss) bx_segment_exception(BX_SS_EXCEPTION, 0);
      else bx_segment_exception(BX_GP_EXCEPTION, 0);
      }

    bx_access_linear((seg->selector.value << 4) + offset,
      length, bx_cpu.cs.selector.rpl, rw, (void *) data);
    }
  /* else protected mode */
  else {
    bx_error(1, "protected mode unsupported!\n");
    }
}


  void
bx_segment_exception(int vector, Bit16u error_code)
{
  bx_error(1, "segment exception!\n");
}


  void
bx_access_linear(Bit32u laddress, Bit8u length, Bit8u pl, 
    Bit8u rw, void *data)
{
  Bit32u paddress1, paddress2;
  Bit8u  len1, len2;

  if ( ((laddress % 4096) + length) > 4096) {
    /* data spans multiple pages */
#ifdef LITTLE_ENDIAN
    len1 = 4096 - (laddress % 4096);
    paddress1 = bx_translate_linear(laddress, pl, rw);

    len2 = length - len1;
    paddress2 = bx_translate_linear(laddress + len1, pl, rw);

    bx_access_physical(paddress1, len1, rw, data);
    bx_access_physical(paddress2, len2, rw, data + len1);

#else /* BIG_ENDIAN */
    len2 = 4096 - (laddress % 4096);
    paddress2 = bx_translate_linear(laddress, pl, rw);

    len1 = length - len2;
    paddress1 = bx_translate_linear(laddress + len2, pl, rw);

    bx_access_physical(paddress1, len1, rw, ((Bit8u *)data) + len2);
    bx_access_physical(paddress2, len2, rw, data);
#endif

    }
  else {
    paddress1 = bx_translate_linear(laddress, pl, rw);
    bx_access_physical(paddress1, length, rw, data);
    }
}



  Bit32u
bx_translate_linear(Bit32u addr, Bit8u pl, Bit8u rw)
{
  if ( !bx_cpu.cr0.pg )
    return(addr);
  else {
    bx_error(1, "paging not supported!\n");
    return(0);  /* keep compiler from complaining */
    }
}



  void
bx_access_physical(Bit32u addr, Bit8u len, Bit8u rw, void *data)
{
  Bit32u page_index, offset;
  Bit32u i, j;
  Bit8u  *data_ptr;


#ifdef LITTLE_ENDIAN
  data_ptr = (Bit8u *) data;

  for (i=1; i<=len; i++, addr++, data_ptr++) {
#else /* BIG_ENDIAN */
  data_ptr = (Bit8u *) data + sizeof(Bit8u) * (len - 1);

  for (i=1; i<=len; i++, addr++, data_ptr--) {
#endif

    page_index = addr / BX_PAGE_SIZE;
    offset     = addr % BX_PAGE_SIZE;

    if ( page_index >= BX_NUM_PAGES )
      bx_error(1, "access beyond physical memory boundary!\n");

    if (bx_phy_mem_pages[page_index].flags) {
      if (bx_phy_mem_pages[page_index].flags & BX_PAGE_UNINITIALIZED) {
	bx_phy_mem_pages[page_index].mem =
          (Bit8u *) malloc((size_t) BX_PAGE_SIZE);
        if (!bx_phy_mem_pages[page_index].mem)
          bx_error(1, "malloc failed allocating memory for physical memory emulation!\n");
        for (j=0; j<BX_PAGE_SIZE; j++)
          bx_phy_mem_pages[page_index].mem[j] = 0;
        bx_phy_mem_pages[page_index].flags -= BX_PAGE_UNINITIALIZED;
        }
      }

    if (rw == BX_READ)
      *(Bit8u *) data_ptr = bx_phy_mem_pages[page_index].mem[offset];
    else
      bx_phy_mem_pages[page_index].mem[offset] = *(Bit8u *) data_ptr;

    } /* for (i=... */
}
