/*
Copyright Notice
================
BOCHS is Copyright 1994 by Kevin P. Lawton.

BOCHS is shareware for PERSONAL USE only.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@tiac.net
*/






#include "bx_bochs.h"



  void
bx_MOVSB_XbYb()
{
  bx_segment_reg_t *seg;
  Bit32u esi, edi, ecx;
  Bit16u si, di;
  Bit8u temp8;

  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.ds;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    esi = si;
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix) {
    ecx = bx_cpu.ecx;
    }

  for (;ecx != 0; ecx--) {
    bx_access_virtual(seg, esi, 1, BX_READ, &temp8);
    bx_access_virtual(&bx_cpu.es, edi, 1, BX_WRITE, &temp8);
    if (bx_cpu.eflags.df) {
      /* decrement ESI */
      bx_cpu.esi--;
      }
    else {
      /* increment ESI */
      bx_cpu.esi++;
      }
    }

  if (bx_rep_prefix) {
    bx_cpu.ecx = 0;
    }
}

  void
bx_MOVSW_XvYv()
{
  bx_segment_reg_t *seg;
  Bit32u esi, edi, temp32, ecx;
  Bit16u si, di, temp16;


  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.ds;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    esi = si;
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix) {
    ecx = bx_cpu.ecx;
    }

  if (bx_32bit_opsize) {
    for (;ecx != 0; ecx--) {
      bx_access_virtual(seg, esi, 4, BX_READ, &temp32);
      bx_access_virtual(&bx_cpu.es, edi, 4, BX_WRITE, &temp32);

      if (bx_cpu.eflags.df) {
        /* decrement ESI */
        bx_cpu.esi -= 4;
        }
      else {
        /* increment ESI */
        bx_cpu.esi += 4;
        }
      }
    } /* if (bx_32bit_opsize) ... */
  else { /* 16 bit opsize mode */
    for (;ecx != 0; ecx--) {
      bx_access_virtual(seg, esi, 2, BX_READ, &temp16);
      bx_access_virtual(&bx_cpu.es, edi, 2, BX_WRITE, &temp16);

      if (bx_cpu.eflags.df) {
        /* decrement ESI */
        bx_cpu.esi -= 2;
        }
      else {
        /* increment ESI */
        bx_cpu.esi += 2;
        }
      }
    }


  if (bx_rep_prefix) {
    bx_cpu.ecx = 0;
    }
}

  void
bx_CMPSB_XbYb()
{
  bx_segment_reg_t *seg;
  Bit32u esi, edi, ecx;
  Bit16u si, di;
  Bit8u src8, dst8, src_b7, dst_b7, diff_8, diff_b7;


  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.ds;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    esi = si;
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix || bx_repne_prefix) {
    ecx = bx_cpu.ecx;
    }

  for (;ecx != 0;) {
    ecx--;
    bx_access_virtual(seg, esi, 1, BX_READ, &src8);
    bx_access_virtual(&bx_cpu.es, edi, 1, BX_READ, &dst8);

    diff_8 = src8 - dst8;

    /* set eflags:
     * CMPSB affects the following flags: O,S,Z,A,P,C
     */
    dst_b7  = dst8  & 0x80;
    src_b7  = src8  & 0x80;
    diff_b7 = diff_8 & 0x80;

    bx_cpu.eflags.of =  (dst_b7 != src_b7) && (diff_b7 != src_b7);
    bx_cpu.eflags.sf = (diff_b7 != 0);
    bx_cpu.eflags.zf = (diff_8 == 0);
    bx_cpu.eflags.af = ((src8 & 0x0F) < (dst8 & 0x0F));
    bx_cpu.eflags.pf = bx_parity_lookup[diff_8];
    bx_cpu.eflags.cf = (src8 < dst8);


    if (bx_cpu.eflags.df) {
      /* decrement ESI */
      bx_cpu.esi--;
      bx_cpu.edi--;
      }
    else {
      /* increment ESI */
      bx_cpu.esi++;
      bx_cpu.edi++;
      }

    if (bx_rep_prefix && bx_cpu.eflags.zf==0) break;
    if (bx_repne_prefix && bx_cpu.eflags.zf!=0) break;
    }

  if (bx_rep_prefix || bx_repne_prefix) {
    bx_cpu.ecx = ecx;
    }
  bx_cpu.edi = edi;
  bx_cpu.esi = esi;
}

  void
bx_CMPSW_XvYv()
{
  bx_segment_reg_t *seg;
  Bit32u esi, edi, ecx;
  Bit16u si, di;
  Bit32u src32, dst32, src_b31, dst_b31, diff_32, diff_b31;
  Bit16u src16, dst16, src_b15, dst_b15, diff_16, diff_b15;


  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.ds;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    esi = si;
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix || bx_repne_prefix) {
    ecx = bx_cpu.ecx;
    }

  for (;ecx != 0;) {
    ecx--;

    if (bx_32bit_opsize) {
      bx_access_virtual(seg, esi, 4, BX_READ, &src32);
      bx_access_virtual(&bx_cpu.es, edi, 4, BX_READ, &dst32);

      diff_32 = src32 - dst32;

      /* set eflags:
       * CMPSB affects the following flags: O,S,Z,A,P,C
       */
      dst_b31  = dst32  & 0x80000000;
      src_b31  = src32  & 0x80000000;
      diff_b31 = diff_32 & 0x80000000;

      bx_cpu.eflags.of =  (dst_b31 != src_b31) && (diff_b31 != src_b31);
      bx_cpu.eflags.sf = (diff_b31 != 0);
      bx_cpu.eflags.zf = (diff_32 == 0);
      bx_cpu.eflags.af = ((src32 & 0x0F) < (dst32 & 0x0F));
      bx_cpu.eflags.pf = bx_parity_lookup[diff_32 & 0xFF];
      bx_cpu.eflags.cf = (src32 < dst32);
      if (bx_cpu.eflags.df) {
	/* decrement ESI */
	bx_cpu.esi -= 4;
	bx_cpu.edi -= 4;
	}
      else {
	/* increment ESI */
	bx_cpu.esi += 4;
	bx_cpu.edi += 4;
	}
      }
    else { /* 16 bit opsize */
      bx_access_virtual(seg, esi, 2, BX_READ, &src16);
      bx_access_virtual(&bx_cpu.es, edi, 2, BX_READ, &dst16);

      diff_16 = src16 - dst16;

      /* set eflags:
       * CMPSB affects the following flags: O,S,Z,A,P,C
       */
      dst_b15  = dst16  & 0x8000;
      src_b15  = src16  & 0x8000;
      diff_b15 = diff_16 & 0x8000;

      bx_cpu.eflags.of =  (dst_b15 != src_b15) && (diff_b15 != src_b15);
      bx_cpu.eflags.sf = (diff_b15 != 0);
      bx_cpu.eflags.zf = (diff_16 == 0);
      bx_cpu.eflags.af = ((src16 & 0x0F) < (dst16 & 0x0F));
      bx_cpu.eflags.pf = bx_parity_lookup[diff_16 & 0xFF];
      bx_cpu.eflags.cf = (src16 < dst16);
      if (bx_cpu.eflags.df) {
	/* decrement ESI */
	bx_cpu.esi -= 2;
	bx_cpu.edi -= 2;
	}
      else {
	/* increment ESI */
	bx_cpu.esi += 2;
	bx_cpu.edi += 2;
	}
      }


    if (bx_rep_prefix && bx_cpu.eflags.zf==0) break;
    if (bx_repne_prefix && bx_cpu.eflags.zf!=0) break;
    }

  if (bx_rep_prefix || bx_repne_prefix) {
    bx_cpu.ecx = ecx;
    }
  bx_cpu.edi = edi;
  bx_cpu.esi = esi;
}

  void
bx_SCASB_ALXb()
{
  Bit32u edi, ecx;
  Bit16u di;
  Bit8u al, dst8, src_b7, dst_b7, diff_8, diff_b7;


  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix || bx_repne_prefix) {
    ecx = bx_cpu.ecx;
    }

  al = bx_cpu.eax;

  for (;ecx != 0;) {
    ecx--;
    bx_access_virtual(&bx_cpu.es, edi, 1, BX_READ, &dst8);

    diff_8 = al - dst8;

    /* set eflags:
     * CMPSB affects the following flags: O,S,Z,A,P,C
     */
    dst_b7  = dst8  & 0x80;
    src_b7  = al  & 0x80;
    diff_b7 = diff_8 & 0x80;

    bx_cpu.eflags.of =  (dst_b7 != src_b7) && (diff_b7 != src_b7);
    bx_cpu.eflags.sf = (diff_b7 != 0);
    bx_cpu.eflags.zf = (diff_8 == 0);
    bx_cpu.eflags.af = ((al & 0x0F) < (dst8 & 0x0F));
    bx_cpu.eflags.pf = bx_parity_lookup[diff_8];
    bx_cpu.eflags.cf = (al < dst8);


    if (bx_cpu.eflags.df) {
      /* decrement ESI */
      bx_cpu.edi--;
      }
    else {
      /* increment ESI */
      bx_cpu.edi++;
      }

    if (bx_rep_prefix && bx_cpu.eflags.zf==0) break;
    if (bx_repne_prefix && bx_cpu.eflags.zf!=0) break;
    }

  if (bx_rep_prefix || bx_repne_prefix) {
    bx_cpu.ecx = ecx;
    }
  bx_cpu.edi = edi;
}

  void
bx_SCASW_eAXXv()
{

  Bit32u edi, ecx;
  Bit16u di;
  Bit32u src32, dst32, src_b31, dst_b31, diff_32, diff_b31;
  Bit16u src16, dst16, src_b15, dst_b15, diff_16, diff_b15;


  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix || bx_repne_prefix) {
    ecx = bx_cpu.ecx;
    }
  if (bx_32bit_opsize) {
    src32 = bx_cpu.eax;
    }
  else /* 16 bit opsize mode */
    src16 = bx_cpu.eax;

  for (;ecx != 0;) {
    ecx--;

    if (bx_32bit_opsize) {
      bx_access_virtual(&bx_cpu.es, edi, 4, BX_READ, &dst32);

      diff_32 = src32 - dst32;

      /* set eflags:
       * CMPSB affects the following flags: O,S,Z,A,P,C
       */
      dst_b31  = dst32  & 0x80000000;
      src_b31  = src32  & 0x80000000;
      diff_b31 = diff_32 & 0x80000000;

      bx_cpu.eflags.of =  (dst_b31 != src_b31) && (diff_b31 != src_b31);
      bx_cpu.eflags.sf = (diff_b31 != 0);
      bx_cpu.eflags.zf = (diff_32 == 0);
      bx_cpu.eflags.af = ((src32 & 0x0F) < (dst32 & 0x0F));
      bx_cpu.eflags.pf = bx_parity_lookup[diff_32 & 0xFF];
      bx_cpu.eflags.cf = (src32 < dst32);

      if (bx_cpu.eflags.df) {
	/* decrement ESI */
	bx_cpu.edi -= 4;
	}
      else {
	/* increment ESI */
	bx_cpu.edi += 4;
	}
      }
    else { /* 16 bit opsize */
      bx_access_virtual(&bx_cpu.es, edi, 2, BX_READ, &dst16);

      diff_16 = src16 - dst16;

      /* set eflags:
       * CMPSB affects the following flags: O,S,Z,A,P,C
       */
      dst_b15  = dst16  & 0x8000;
      src_b15  = src16  & 0x8000;
      diff_b15 = diff_16 & 0x8000;

      bx_cpu.eflags.of =  (dst_b15 != src_b15) && (diff_b15 != src_b15);
      bx_cpu.eflags.sf = (diff_b15 != 0);
      bx_cpu.eflags.zf = (diff_16 == 0);
      bx_cpu.eflags.af = ((src16 & 0x0F) < (dst16 & 0x0F));
      bx_cpu.eflags.pf = bx_parity_lookup[diff_16 & 0xFF];
      bx_cpu.eflags.cf = (src16 < dst16);

      if (bx_cpu.eflags.df) {
	/* decrement ESI */
	bx_cpu.edi -= 2;
	}
      else {
	/* increment ESI */
	bx_cpu.edi += 2;
	}
      }

    if (bx_rep_prefix && bx_cpu.eflags.zf==0) break;
    if (bx_repne_prefix && bx_cpu.eflags.zf!=0) break;
    }

  if (bx_rep_prefix || bx_repne_prefix) {
    bx_cpu.ecx = ecx;
    }
  bx_cpu.edi = edi;
}

  void
bx_STOSB_YbAL()
{
  bx_segment_reg_t *seg;
  Bit32u edi, ecx;
  Bit16u di;
  Bit8u al;

  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.es;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix) {
    ecx = bx_cpu.ecx;
    }

  for (;ecx != 0; ecx--) {
    al = bx_cpu.eax & 0xFF;
    bx_access_virtual(seg, edi, 1, BX_WRITE, &al);

    if (bx_cpu.eflags.df) {
      /* decrement ESI */
      bx_cpu.esi--;
      }
    else {
      /* increment ESI */
      bx_cpu.esi++;
      }
    }

  if (bx_rep_prefix) {
    bx_cpu.ecx = 0;
    }
}

  void
bx_STOSW_YveAX()
{
  bx_segment_reg_t *seg;
  Bit32u edi, eax, ecx;
  Bit16u di, ax;


  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.es;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(edi, BX_32BIT_REG_EDI);
    }
  else {
    BX_READ_16BIT_REG(di, BX_16BIT_REG_DI);
    edi = di;
    }

  ecx = 1;
  if (bx_rep_prefix) {
    ecx = bx_cpu.ecx;
    }

  if (bx_32bit_opsize) {
    for (;ecx != 0; ecx--) {
      eax = bx_cpu.eax;
      bx_access_virtual(seg, edi, 4, BX_WRITE, &eax);
      if (bx_cpu.eflags.df) {
        /* decrement ESI */
        bx_cpu.esi -= 4;
        }
      else {
        /* increment ESI */
        bx_cpu.esi += 4;
        }
      }
    } /* if (bx_32bit_opsize) ... */
  else { /* 16 bit opsize mode */
    for (;ecx != 0; ecx--) {
      ax = bx_cpu.eax & 0xFFFF;
      bx_access_virtual(seg, edi, 2, BX_WRITE, &ax);
      if (bx_cpu.eflags.df) {
        /* decrement ESI */
        bx_cpu.esi -= 2;
        }
      else {
        /* increment ESI */
        bx_cpu.esi += 2;
        }
      }
    }


  if (bx_rep_prefix) {
    bx_cpu.ecx = 0;
    }
}


  void
bx_LODSB_ALXb()
{
  bx_segment_reg_t *seg;
  Bit32u esi, ecx;
  Bit16u si;
  Bit8u al;

  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.ds;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
    }
  else {
    BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
    esi = si;
    }

  ecx = 1;
  if (bx_rep_prefix) {
    ecx = bx_cpu.ecx;
    }

  for (;ecx != 0; ecx--) {
    bx_access_virtual(seg, esi, 1, BX_READ, &al);
    bx_cpu.eax = (bx_cpu.eax & 0xFFFFFF00) | al;
    if (bx_cpu.eflags.df) {
      /* decrement ESI */
      bx_cpu.esi--;
      }
    else {
      /* increment ESI */
      bx_cpu.esi++;
      }
    }

  if (bx_rep_prefix) {
    bx_cpu.ecx = 0;
    }
}

  void
bx_LODSW_eAXXv()
{
  bx_segment_reg_t *seg;
  Bit32u esi, eax, ecx;
  Bit16u si, ax;


  if (bx_segment_override) {
    seg = bx_segment_override;
    }
  else {
    seg = &bx_cpu.ds;
    }

  if (bx_32bit_addrsize) {
    BX_READ_32BIT_REG(esi, BX_32BIT_REG_ESI);
    }
  else {
    BX_READ_16BIT_REG(si, BX_16BIT_REG_SI);
    esi = si;
    }

  ecx = 1;
  if (bx_rep_prefix) {
    ecx = bx_cpu.ecx;
    }

  if (bx_32bit_opsize) {
    for (;ecx != 0; ecx--) {
      bx_access_virtual(seg, esi, 4, BX_READ, &eax);
      bx_cpu.eax = eax;
      if (bx_cpu.eflags.df) {
        /* decrement ESI */
        bx_cpu.esi -= 4;
        }
      else {
        /* increment ESI */
        bx_cpu.esi += 4;
        }
      }
    } /* if (bx_32bit_opsize) ... */
  else { /* 16 bit opsize mode */
    for (;ecx != 0; ecx--) {
      bx_access_virtual(seg, esi, 2, BX_READ, &ax);
      bx_cpu.eax = (bx_cpu.eax & 0xFFFF0000) | ax;
      if (bx_cpu.eflags.df) {
        /* decrement ESI */
        bx_cpu.esi -= 2;
        }
      else {
        /* increment ESI */
        bx_cpu.esi += 2;
        }
      }
    }


  if (bx_rep_prefix) {
    bx_cpu.ecx = 0;
    }
}
