/*
Copywrite 1994 by Kevin P. Lawton 

This file is part of the IODEV (Input Output DEVices) component of BOCHS.

The IODEV component is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

The IODEV component is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the IODEV component; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/





#include "../bx_bochs.h"


/* maximum number of emulated devices allowed.  floppy, vga, etc...
   you can increase this to anything below 256 since an 8-bit handle
   is used for each device */
#define BX_MAX_IO_DEVICES 20

/* number of IRQ lines supported.  In an ISA PC there are two
   PIC chips cascaded together.  each has 8 IRQ lines, so there
   should be 16 IRQ's total */
#define BX_MAX_IRQS 16

#define BX_IO_ADDRESS_SPACE        10
#define BX_MEMORY_ADDRESS_SPACE    11

#define BX_CALL_REASON_READ      10
#define BX_CALL_REASON_WRITE     11



#define BX_NO_IRQ  -1
#define BX_NO_FD   -1

#define BX_START_BIOS_ADDR   0x
#define BX_END_BIOS_ADDR     0x

#define BX_FD_EVENT_READ       1
#define BX_FD_EVENT_WRITE      2
#define BX_FD_EVENT_EXCEPTION  4




typedef struct {
  Bit8u     (* io_handler)(Bit8u, Bit8u, Bit32u, Bit8u);
  char    *handler_name;
  Bit32u  start_addr;
  Bit32u  end_addr;
  Bit8u   addr_space;
  int     irq;
  } bx_iodev_t;






void bx_timer_setup(void);

void bx_register_io_handlers(void);
int  bx_register_io_handler( bx_iodev_t );
void bx_register_int_vector(Bit8u interrupt, Bit32u bios_addr,
  void (* funct)(int vector));
int  bx_register_timer( void (*funct)(void), Bit32u useconds);


Bit8u bx_unmapped_io_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_dma_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_pit_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_pic_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_keyboard_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_cmos_ram_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_hard_drive_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_parallel_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_serial_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_floppy_handler(Bit8u, Bit8u, Bit32u, Bit8u);
Bit8u bx_vga_handler(Bit8u, Bit8u, Bit32u, Bit8u);

void bx_sigvtalrm_handler(int sig);
void bx_cpu_loop(void);
void bx_error(int, ...);

void bx_outp(Bit16u addr, Bit8u value);
Bit8u bx_inp(Bit16u addr);

void bx_bios_post(void);

void bx_trigger_irq(int irq_no);

void bx_pit_int_handler(int vector);
void bx_keyboard_int_handler(int vector);
void bx_cmos_rtc_int_handler(int vector);
void bx_hard_drv_int_handler(int vector);
void bx_parallel_int_handler(int vector);
void bx_serial_int_handler(int vector);
void bx_floppy_int_handler(int vector);

Bit8u bx_peek(Bit32u addr);

void bx_poke(Bit32u addr, Bit8u value);

void bx_iodev_setup(void);
void bx_init_pic(void);
void bx_init_pit(void);
void bx_init_dma(void);
void bx_init_keyboard(void);
void bx_init_cmos(void);
void bx_init_hard_drive(void);
void bx_init_parallel(void);
void bx_init_serial(void);
void bx_init_floppy(void);
void bx_init_vga(void);
