/*
Copywrite 1994 by Kevin P. Lawton 

This file is part of the IODEV (Input Output DEVices) component of BOCHS.

The IODEV component is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

The IODEV component is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the IODEV component; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <stdio.h>
#include "iodev.h"



  void
bx_register_io_handlers(void)
{
  bx_iodev_t  device;

  /* 8237 DMA controller */
  device.io_handler   = bx_dma_handler;
  device.handler_name = "DMA controller";
  device.start_addr   = 0x0000;
  device.end_addr     = 0x000F;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = BX_NO_IRQ;
  bx_register_io_handler(device);

  device.start_addr   = 0x0081;
  device.end_addr     = 0x008F;
  bx_register_io_handler(device);

  device.start_addr   = 0x00C0;
  device.end_addr     = 0x00DF;
  bx_register_io_handler(device);


  /* 8254 PIT (Programmable Interval Timer) */
  device.io_handler   = bx_pit_handler;
  device.handler_name = "8254 PIT";
  device.start_addr   = 0x0040;
  device.end_addr     = 0x0043;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = 0;
  bx_register_io_handler(device);


  /* 8259 PIC (Programmable Interrupt Controller) */
  device.io_handler   = bx_pic_handler;
  device.handler_name = "8259 PIC";
  device.start_addr   = 0x0020;
  device.end_addr     = 0x0021;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = BX_NO_IRQ;
  bx_register_io_handler(device);

  device.start_addr = 0x00A0;
  device.end_addr   = 0x00A1;
  bx_register_io_handler(device);


  /* 8042 keyboard controller */
  device.io_handler   = bx_keyboard_handler;
  device.handler_name = "8042 Keyboard controller";
  device.start_addr   = 0x0060;
  device.end_addr     = 0x0060;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = 1;
  bx_register_io_handler(device);

  device.start_addr   = 0x0064;
  device.end_addr     = 0x0064;
  bx_register_io_handler(device);


  /* CMOS RAM & RTC */
  device.io_handler   = bx_cmos_ram_handler;
  device.handler_name = "CMOS RAM";
  device.start_addr   = 0x0070;
  device.end_addr     = 0x0071;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = BX_NO_IRQ;
  bx_register_io_handler(device);


  /* HARD DRIVE 0 */
  device.io_handler   = bx_hard_drive_handler;
  device.handler_name = "Hard Drive 0";
  device.start_addr   = 0x01F0;
  device.end_addr     = 0x01F7;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = 14;
  bx_register_io_handler(device);


  /* PARALLEL PORT 1 */
  device.io_handler   = bx_parallel_handler;
  device.handler_name = "Parallel Port 1";
  device.start_addr   = 0x03BC;
  device.end_addr     = 0x03BE;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = 7;
  bx_register_io_handler(device);


  /* SERIAL PORT 1 */
  device.io_handler   = bx_serial_handler;
  device.handler_name = "Serial Port 1";
  device.start_addr   = 0x03F8;
  device.end_addr     = 0x03FF;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = 4;
  bx_register_io_handler(device);


  /* FLOPPY DRIVE */
  device.io_handler   = bx_floppy_handler;
  device.handler_name = "Floppy Drive";
  device.start_addr   = 0x03F0;
  device.end_addr     = 0x03F7;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = 6;
  bx_register_io_handler(device);


  /* VGA */
  device.io_handler   = bx_vga_handler;
  device.handler_name = "VGA";
  device.start_addr   = 0x03B4;
  device.end_addr     = 0x03B5;
  device.addr_space   = BX_IO_ADDRESS_SPACE;
  device.irq          = BX_NO_IRQ;
  bx_register_io_handler(device);

  device.start_addr   = 0x03BA;
  device.end_addr     = 0x03BA;
  bx_register_io_handler(device);

  device.start_addr   = 0x03C0;
  device.end_addr     = 0x03CF;
  bx_register_io_handler(device);

  device.start_addr   = 0x03D4;
  device.end_addr     = 0x03D5;
  bx_register_io_handler(device);

  device.start_addr   = 0x03DA;
  device.end_addr     = 0x03DA;
  bx_register_io_handler(device);

  device.start_addr   = 0xA0000;
  device.end_addr     = 0xBFFFF;
  device.addr_space   = BX_MEMORY_ADDRESS_SPACE;
  bx_register_io_handler(device);
}

